/* This file is part of the Allegria project
   Copyright (C) 2004 Francois Schiettekatte <francois.schiettekatte@umontreal.ca>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

// Dialog in which the user sets what assemblage of detection channel will be plotted as
// x and y coordinate for each event, and how to calculate the energy.
// Also allows the user to set TOF-related parameters.

#include <qgrid.h>
#include <qlayout.h> 
#include <qhbox.h> 
#include <qlabel.h>
#include <qvalidator.h>
#include "channelSelector.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

channelSelector::channelSelector(long nnchan, SXDisplay ddisplay, int	divide, int specsize, QWidget *parent, QString name )
            : QDialog( parent, name, true )
{

	int i;
	setCaption("Channel Selector");

	nchan = nnchan;
	display = ddisplay;
	//memcpy((void *)&display, (void *)&ddisplay, sizeof(display));

	QString abc("ABCDEFGHIJKLMNOPQRSTUVWXYZ");

	QStringList formula;
	formula << "none" << "A+B+C+D" << "A+B-C-D" << "(A-B)/(C+D)" << "(A+B+C+D)/Ener";

	QStringList expr;
	expr << "X = ";
	for(i=1; i<NUMPARAM; i++) // paprameters to specify
		expr << " , " + abc[i-1] + " = ";

	QStringList channels;
	channels << " ";
	for(i=1; i<=nchan; i++) // user will select channel 0 for no channel
		channels << QString::number(i);

	QBoxLayout *vbox = new QVBoxLayout(this, 2, 2);
	
	// NUMPARAM combobox in each group: formula, and parameters A, B, C, D

	vbox->addWidget(new QLabel("Display (for elements identification):",this));

	// X
	QHBox *hboxX = new QHBox(this);
	vbox->addWidget(hboxX);

	for(i=0; i<NUMPARAM; i++) {
		new QLabel(expr[i],hboxX);
		X[i] = new QComboBox(hboxX);
		X[i]->insertStringList(i?channels:formula);
		X[i]->setCurrentItem(display.paramX[i]);
	}

	// Y
	QHBox *hboxY = new QHBox(this);
	vbox->addWidget(hboxY);
	expr[0] = "Y = ";

	for(i=0; i<NUMPARAM; i++) {
		new QLabel(expr[i],hboxY);
		Y[i] = new QComboBox(hboxY);
		Y[i]->insertStringList(i?channels:formula);
		Y[i]->setCurrentItem(display.paramY[i]);
	}

	// E
	vbox->addWidget(new QLabel("       ",this) ); // spacer
	vbox->addWidget(new QLabel("Energy calculation for projection:",this));
	QHBox *hboxE = new QHBox(this);
	vbox->addWidget(hboxE);
	expr[0] = "E = ";

	for(i=0; i<NUMPARAM; i++) {
		new QLabel(expr[i],hboxE);
		E[i] = new QComboBox(hboxE);
		E[i]->insertStringList(i?channels:formula);
		E[i]->setCurrentItem(display.paramE[i]);
	}

		
	// S
	vbox->addWidget(new QLabel("       ",this) ); // spacer
	vbox->addWidget( new QLabel("Spectrum directly from channel:",this) );
	QHBox *hboxS = new QHBox(this);
	vbox->addWidget(hboxS);
	expr[0] = "S = ";

	for(i=0; i<NUMPARAM; i++) {
		new QLabel(expr[i],hboxS);
		S[i] = new QComboBox(hboxS);
		S[i]->insertStringList(i?channels:formula);
		S[i]->setCurrentItem(display.paramS[i]);
	}
		
		
	// TOF
	vbox->addWidget(new QLabel("       ",this) ); // spacer
	vbox->addWidget( new QLabel("TOF reshape parameters (set Camera length to 0 to avoid reshape)",this) );
	QHBox *hboxTOF = new QHBox(this);
	vbox->addWidget(hboxTOF);

    QDoubleValidator * validatorD = new QDoubleValidator(this);
	new QLabel("  Camera length :  ",hboxTOF);
    camlengthEdit = new QLineEdit( hboxTOF );
    camlengthEdit->setText(QString::number(display.caml));
    camlengthEdit->setValidator(validatorD);
	//camlengthEdit->setFixedWidth(50);

	new QLabel("     delay :  ",hboxTOF);
    delayEdit = new QLineEdit( hboxTOF );
    delayEdit->setText(QString::number(display.delay));
    delayEdit->setValidator(validatorD);
	//delayEdit->setFixedWidth(50);

	new QLabel("     E0 :  ",hboxTOF);
    E0Edit = new QLineEdit( hboxTOF );
    E0Edit->setText(QString::number(display.E0));
    E0Edit->setValidator(validatorD);
	//E0Edit->setFixedWidth(50);

    connect(camlengthEdit, SIGNAL(textChanged(const QString &)), this, SLOT(ableTOFedit()) );
	ableTOFedit();


	vbox->addWidget(new QLabel("       ",this) ); // spacer

	QHBox *hbox1 = new QHBox(this);
	vbox->addWidget(hbox1);

	new QLabel("spectrum size: ",hbox1);
    spectrumSize = new QLineEdit( hbox1 );
    spectrumSize->setText(QString::number(specsize));
    QIntValidator * validator = new QIntValidator(0,5, this);
    spectrumSize->setValidator(validator);
	spectrumSize->setFixedWidth(50);

	new QLabel("division factor:",hbox1);
    spectrumDivide = new QLineEdit( hbox1 );
    spectrumDivide->setText(QString::number(divide));
    QIntValidator * validator2 = new QIntValidator(0,2, this);
    spectrumDivide->setValidator(validator2);
	spectrumDivide->setFixedWidth(50);

    QHBox *hboxOk = new QHBox( this );
	vbox->addWidget(hboxOk);

	new QLabel("       ",hboxOk); // spacer
    ok = new QPushButton( "OK", hboxOk );
    ok->setDefault(true);
    connect( ok, SIGNAL(clicked()), SLOT(accept()) );

    cancel = new QPushButton( "Cancel", hboxOk );
    connect( cancel, SIGNAL(clicked()), SLOT(reject()) );

	//vbox->resize( minimumSizeHint());
}

channelSelector::~channelSelector()
{

}

SXDisplay channelSelector::getDisplay()
{
	for(int i=0; i<NUMPARAM; i++) {
		display.paramX[i] = X[i]->currentItem();
		display.paramY[i] = Y[i]->currentItem();
		display.paramE[i] = E[i]->currentItem();
		display.paramS[i] = S[i]->currentItem();
	}
	display.caml  = camlengthEdit->text().toDouble();
	display.delay = delayEdit->text().toDouble();
	display.E0    = E0Edit->text().toDouble();

	return display;

}

void channelSelector::ableTOFedit() {

	if(camlengthEdit->text().toDouble()==0.) {
		delayEdit->setEnabled(false);
		E0Edit->setEnabled(false);
	} else {
		delayEdit->setEnabled(true);
		E0Edit->setEnabled(true);
	}

}