/* This file is part of the Allegria project
   Copyright (C) 2004 Francois Schiettekatte <francois.schiettekatte@umontreal.ca>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

// Application window that allows the user to plot list-mode events
// in a versatile way in order to extract its different components 

#include <qdir.h>
#include <qmenubar.h>
#include <qtoolbar.h>
#include <qtoolbutton.h>
#include <qworkspace.h>
#include <qfiledialog.h>
#include <qprogressdialog.h>
#include <qmessagebox.h>
#include <qimage.h>
#include <qregexp.h>
#include <qinputdialog.h>
#include <stdlib.h>

#include "spectrumextractorwindow.h"
#include "../XYQt/JamaQRDecomposition.h"
#include "../XYQt/colorseries.h"
#include "pnpoly.h"

#include "SpectrumExtractor.xpm"
#include "fileopen.xpm"
#include "filesave.xpm"
#include "filesavemulti.xpm"
#include "channel.xpm"
#include "projection.xpm"
#include "display.xpm"
#include "reshape.xpm"
#include "polygon.xpm"

#define MAXDELAY 12000  // a value (mush) larger than the delay in the timing signal (channel units)

SpectrumExtractorWindow::SpectrumExtractorWindow( QWidget* parent, const char* name, WFlags f )
                        : QMainWindow( parent, name, f )
{
    long i, j;
    int posx, posy, sizex, sizey;

    setCaption("Spectrum Extractor");
    setIcon( QPixmap(spectrumExtractor_xpm) );

    ndata = 0;
    nparamerror = true;

    for(i = 0; i<MAXCHAN+1; i++) data[i] = NULL;

    // read some working parameters saved after last run
    char s[1000];
    ifstream fi("spectrumextractor.ini");
    if(fi) {
        fi.getline(s, 1000);
        workingDir = s;
        fi >> specsize >> divide;
        fi.getline(s, 1000); // get end of line

        fi.getline(s, 1000); // window geometry
        fi >> posx >> posy >> sizex >> sizey;
        fi.getline(s, 1000); // get end of line

        fi.getline(s, 1000); // "display parameters"
        for(j = 0; j<NUMDISPLAY; j++) {
            fi >> display[j].caml >> display[j].delay >> display[j].E0;
            for(i = 0; i<NUMPARAM; i++)  {
                fi >> display[j].paramX[i] >> display[j].paramY[i] >> display[j].paramE[i] >> display[j].paramS[i];
            }
            fi >> display[j].xmin >> display[j].xmax >> display[j].ymin >> display[j].ymax;

		    //display name
            fi.getline(s, 1000);
            display[j].name = s;
            display[j].name = display[j].name.simplifyWhiteSpace();
        }

        fi.getline(s, 1000); // "polygons"
        for(j = 0; j<MAXELEMENT; j++) {
            int element, npolygon;
            fi >> element >> npolygon;
            if(npolygon) {
                // if defined, read and store mass definition polygon in a dataset
                polygon[j] = new FSDataSet(QString::number(j));
                polygon[j]->linePen = QPen(colorSeries(j),1);
                polygon[j]->setsymbol(0);


                double x, y;
                for(i = 0; i<npolygon; i++)  {
                    fi >> x >> y;
                    polygon[j]->append(x,y);
                }
            } else {
                polygon[j] = NULL;
            }
        }

    } else {
        // defaults
        QWidget *d = QApplication::desktop();
        posx=0; posy=0; sizex=d->width(); sizey=d->height();

        workingDir = QDir::currentDirPath();
        specsize = 512;
        divide = 8;

        for(j = 0; j<NUMDISPLAY; j++) {
            display[j].caml = 0.;
            display[j].delay = 0.;
            display[j].E0 = 0.;
            display[j].xmin = 0.;
            display[j].xmax = 10.;
            display[j].ymin = 0.;
            display[j].ymax = 10.;
            display[j].name = "display &"+QString::number(j);

            for(i = 0; i<NUMPARAM; i++) {
                display[j].paramX[i] = 0;
                display[j].paramY[i] = 0;
                display[j].paramE[i] = 0;
                display[j].paramS[i] = 0;
            }
            display[j].paramX[0] = 1; // by default, plot first channel as x ...
            display[j].paramX[1] = 1;
            display[j].paramY[0] = 1; // ... and second channel as y
            display[j].paramY[1] = 2;
            display[j].paramE[0] = 1; // first channel being the energy
            display[j].paramE[1] = 1;
        }

        for(j = 0; j<MAXELEMENT; j++) {
            polygon[j] = NULL; // no polygon defined
        }

    }
    fi.close();

    // graphs container (MDI management)
    QWorkspace* main = new QWorkspace(this);

    // menu setup
    QPopupMenu *file = new QPopupMenu( this );
    file->insertItem( "&Open event file",  this, SLOT(openEventFile()), CTRL+Key_O );
    file->insertItem( "&Write energy spectra in a single file", this, SLOT(writeMultiColumn()));
    file->insertItem( "Write energy spectra in &separate files", this, SLOT(writeMultiFile()));
    file->insertItem( "Write selected element events in separate files", this, SLOT(writeElementEventFile()));
    file->insertSeparator();    
	file->insertItem( "E&xit",  qApp, SLOT(quit()), CTRL+Key_Q );
	file->insertItem( "&About", this, SLOT(about()));

    recall = new QPopupMenu( this );
    for(i=1; i<NUMDISPLAY; i++)
        recall->setItemParameter(recall->insertItem( display[i].name,  this, SLOT(recallDisplay(int))),i);
    store = new QPopupMenu( this );
    for(i=1; i<NUMDISPLAY; i++)
        store->setItemParameter(store->insertItem( display[i].name,  this, SLOT(storeDisplay(int))),i);
    QPopupMenu *displaymenu = new QPopupMenu( this );
    displaymenu->insertItem( "Cha&nnel selector", this, SLOT(selectChannel()), CTRL+Key_N );
    displaymenu->insertItem( "&Apply reshape", this, SLOT(changeShape()), CTRL+Key_A );
    displaymenu->insertItem( "&Cancel reshape", this, SLOT(cancelReshape()), CTRL+Key_Z );
    displaymenu->insertItem( "&Recall display", recall );
    displaymenu->insertItem( "&Store current display as", store );
    displaymenu->insertItem( "Define &polygon", this, SLOT(definePolygon()), CTRL+Key_P );

    QPopupMenu *mass = new QPopupMenu( this );
    mass->setItemParameter(mass->insertItem( "&Energy Projection", this, SLOT(Eprojection()), CTRL+Key_E ),1); // type 1
    mass->setItemParameter(mass->insertItem( "Energy Projection using &time && reshape parameters", this, SLOT(projection(int)), CTRL+Key_T ),2); // type 2
    mass->insertItem( "Projection on the &Y axis", this, SLOT(Yprojection()), CTRL+Key_Y );
    mass->setItemParameter(mass->insertItem( "&Count rate progression", this, SLOT(projection(int)), CTRL+Key_C ),3); // type 3

    QPopupMenu *window = new QPopupMenu( this );
    window->insertItem( "&Tile",  main, SLOT(tile()), CTRL+SHIFT+Key_T );
    window->insertItem( "&Cascade",  main, SLOT(cascade ()), CTRL+SHIFT+Key_C );

    menu = new QMenuBar( this );
    menu->insertItem( "&File", file );
    menu->insertItem( "&Display", displaymenu );
    menu->insertItem( "&Projections", mass );
    menu->insertItem( "&Window", window );

    // toolbar
    setUsesBigPixmaps(true);
    QToolBar * tbar = new QToolBar("toolbar", this);
    new QToolButton(QPixmap(fileopen), "Open event file", "Open event file", this, SLOT(openEventFile()), tbar );
    new QToolButton(QPixmap(filesave), "Write energy spectra in a single file", "saveSingleFile", this, SLOT(writeMultiColumn()), tbar );
    new QToolButton(QPixmap(filesavemulti_xpm), "Write energy spectra in separate files", "saveSeparateFiles", this, SLOT(writeMultiFile()), tbar );
    tbar->addSeparator();
    new QToolButton(QPixmap(channel_xpm), "channel selector", "channel", this, SLOT(selectChannel()), tbar );
    QToolButton *rec = new QToolButton(QPixmap(display_xpm), "recall display", "display", this, NULL , tbar );
    rec->setPopup(recall);
    rec->setPopupDelay(0);

    new QToolButton(QPixmap(reshape_xpm), "apply reshape", "reshape", this, SLOT(changeShape()), tbar );
    new QToolButton(QPixmap(polygon_xpm), "define polygon", "polygon", this, SLOT(definePolygon()), tbar );
    tbar->addSeparator();
    new QToolButton(QPixmap(projection_xpm), "Energy Projection", "Energy Projection", this, SLOT(Eprojection()), tbar );

    // graphs
    eventGraph   = new FSGraph("events_graph.agr",main);
    //eventGraph->x->title = QString("x");
    //eventGraph->y->title = QString("y");
    //eventGraph->setlegendPos(1);
    eventGraph->setGraphName("events graph");
    eventGraph->readOnlyGraph = false;
    eventGraph->allowSpots = true;
    eventGraph->zoom(display[0].xmin, display[0].xmax, display[0].ymin, display[0].ymax);
    energyGraph  = new FSGraph("mass_sep_ener_spec.agr",main);
    //energyGraph->x->title = QString("energy channel");
    //energyGraph->y->title = QString("counts");
    //energyGraph->setlegendPos(2);
    energyGraph->setGraphName("Mass separated energy spectra");

    // A widget...
    periodicTableWidget = new QWidget(main, "periodicTableWidget");
    periodicTableWidget->setBaseSize(QSize(sizex/2.1,sizey/2.4));
    periodicTableWidget->setCaption("Mass to be extracted");
    // ...that will containt a periodic table widget to select a single element
    periodicTable = new PeriodicTable(sizex/2.1,sizey/2.4, false, true, periodicTableWidget);
    //periodicTable->setSelected(selectedAtom,true);
    connect(periodicTable, SIGNAL(elementSelected()), this, SLOT(updateData()) );

    // name polygon datasets according to their Z value
    for(long jj = 0; jj<MAXELEMENT; jj++)
        if(polygon[jj])
            polygon[jj]->legend = periodicTable->returnAtom(polygon[jj]->legend.toInt()).name;

    setCentralWidget(main);
    move(posx, posy);
    resize(sizex,sizey);

    qApp->processEvents();     // process events so the main window shows up and we can tile its children
    main->tile();

}

SpectrumExtractorWindow::~SpectrumExtractorWindow()
{
    long j;
    display[0].xmin = eventGraph->x->minC;
    display[0].xmax = eventGraph->x->maxC;
    display[0].ymin = eventGraph->y->minC;
    display[0].ymax = eventGraph->y->maxC;

    if(ndata)
        for(j=0; j<nchan; j++)
            delete data[j];

    ndata = 0;

    // save status (including reshape params and polygons)
    ofstream fo("spectrumextractor.ini");
    fo << workingDir << "\n"
       << specsize << " " << divide << "\n";
    fo << "window geometry" << "\n";
    fo << x() << " " << y() << " " << width() << " " << height() << "\n";
    fo << "display parameters" << "\n";
    for(j = 0; j<NUMDISPLAY; j++) {
        fo << display[j].caml << " " << display[j].delay << " " << display[j].E0 << " ";
        for(long i = 0; i<NUMPARAM; i++)  {
            fo << display[j].paramX[i] << " ";
            fo << display[j].paramY[i] << " ";
            fo << display[j].paramE[i] << " ";
            fo << display[j].paramS[i] << " ";
        }
        fo << display[j].xmin << " " << display[j].xmax << " "
           << display[j].ymin << " " << display[j].ymax << " "
           << display[j].name.latin1() << "\n";
    }

    fo << "polygons" << "\n";
    for(j = 0; j<MAXELEMENT; j++) {
        int npolygon = polygon[j]?polygon[j]->getndata():0;
        fo << j << " " << npolygon  << " " ;
        if(npolygon) {
            for(long i = 0; i<npolygon; i++)  {
                fo << polygon[j]->xp[i] << " " << polygon[j]->yp[i] << " ";
            }
        }
        fo << "\n";
    }
}

// reshape event graph so spots are horizontal (representing masses in the case of ToF)
void SpectrumExtractorWindow::changeShape()
{

    if(eventGraph->spots->getndata()<3) {
        QMessageBox::warning( this, "Spectrum Extractor", "A reshape requires 3 marks.", QMessageBox::Ok, 0);
        return;
    }

    if(eventGraph->spots->getndata()>3) {
        switch( QMessageBox::warning( this, "Spectrum Extractor", "A reshape requires 3 marks. Should we use only the first 3 marks?",
            "Yes", "No", "Cancel", 0, 2 ) ) {
        case 0: //yes
            break;
        case 1: // No
            eventGraph->spots->clear();
            // no brake, we want to return
        case 2: // cancel
        default:
            return;
        }
    }

    Atom atom = selectElement("Select element that has been marked");
    double m = atom.mass;

    double caml  = display[0].caml;
    double delay = display[0].delay;
    double E0    = display[0].E0;


        double Ea = eventGraph->spots->xp[0];
        double Eb = eventGraph->spots->xp[1];
        double Ec = eventGraph->spots->xp[2];

        // these are masses if graph currently reshaped
        double ta = eventGraph->spots->yp[0];
        double tb = eventGraph->spots->yp[1];
        double tc = eventGraph->spots->yp[2];

        if(caml!=0.) {
            // graph is reshaped and we need de-shape time values (i.e. time and not mass)
            ta = delay - pow(ta*caml*caml/2/(Ea+E0),0.5);
            tb = delay - pow(tb*caml*caml/2/(Eb+E0),0.5);
            tc = delay - pow(tc*caml*caml/2/(Ec+E0),0.5);
        }

        // scan on delay values to find which will make the 3 spots colinear
        // (here's the trick: we have an equation of the form (E+E0) = m l^2/2(t0-t)^2
        // where t0 is the delay and l the camera length. This rewrites E = m l^2 x/2 - E0
        // with x = 1/(t0-t)^2. If we find a value of t0 making the 3 points colinear
        // when we plot E(i) vs x(t(i)) in this equation, we can deduce E0 and l form the
        // slope and intercept)

        int jmin = 0, jmax = MAXDELAY;
        double slopeDiff = 1.e300;
        for(int j=0; j<jmax; j++){
            // ya, yb, yc represent actually 1/x relatively to the above discussion
            double ya = (j-ta)*(j-ta);
            double yb = (j-tb)*(j-tb);
            double yc = (j-tc)*(j-tc);
            if(ya!=0. && yb!=0. && yc!=0.) {
                // we actually compute the difference of inverse slope
                // (which should also be a minimum) to avoid some side effect when y~0. i.e. j~t(a,b,c)
                double dsl = fabs((1/ya-1/yb)/(Ea-Eb)-(1/ya-1/yc)/(Ea-Ec));
                if( dsl<slopeDiff ) {
                    slopeDiff = dsl;
                    jmin = j;
                }
            }
        }

        double ya = (jmin-ta)*(jmin-ta);
        double yb = (jmin-tb)*(jmin-tb);
        double yc = (jmin-tc)*(jmin-tc);

        if(ya==0. || yb==0. || yc==0.) return; // something went wrong!!

        double caml2 = 2*(Ea-Eb)/(1/ya-1/yb)/m;  // slope = m l^2/2 => l^2 = 2*slope/m
        if(caml2<0) return; // something went wrong!!

        // set values for selected mass
        display[0].caml  = pow(caml2,0.5);
        display[0].delay = jmin;
        display[0].E0    = m*caml2/ya/2.-Ea;

        updateData();

    eventGraph->spots->clear();

}

// open a "standard" event file
bool SpectrumExtractorWindow::openEventFile()
{

    // get file name and path
    QString fileName( QFileDialog::getOpenFileName(workingDir, "Allegria Event Files (*.dat);;All Files (*.*)") );
    if ( fileName.isNull() ) return false; // no file selected
    // extract file name (remove path and extension)
    workingName = fileName;
    int slash = fileName.findRev(QRegExp("[\\/\\\\]"));
    if(slash!=-1) {
        workingName = fileName.right( fileName.length()-slash-1 );
        workingDir = fileName.left( slash );
    }
    int pos = workingName.findRev(".");
    if(pos>-1) workingName = workingName.left(pos);

    // open file
    ifstream f(fileName);
    if(!f) return false;    // didn't succeed opening file

    // check header
    char ss[1000];
    f.getline(ss,1000);
    if( strncmp(ss,"Allegria event file", 18)!=0 && strncmp(ss,"Alegria event file", 18)!=0) { // 2nd one for compatibility with early versions
        QMessageBox::warning( this, "Spectrum Extractor", "Not an Allegria event file", QMessageBox::Ok, 0);
        f.close();
        return false;
    }

    // read number of events and channels
    f >> ndata >> nchan;
    if(ndata<=0 || nchan<=0 || nchan>MAXCHAN) {
        QMessageBox::warning( this, "Spectrum Extractor", "Wrong number of events or channels in file header", QMessageBox::Ok, 0);
        f.close();
        return false;
    }

    // allocate space
    long i, j;
    for(j=0; j<nchan; j++)
        data[j] = new short[ndata];

    // setup a progress bar (this is a modal dialog so we need to call qApp->processEvents())
    QProgressDialog progress( "Loading " + QString::number(ndata) + " events", "Abort", ndata,                                              this, "progress", TRUE );
    progress.setCaption("Progress");

    // read the data
    for(i=0; i<ndata; i++)  {
        for(j=0; j<nchan; j++)  {
            f >> data[j][i];
        }

        // every 1%, update the progress bar
        if(i%(ndata/100+1)==0) {
            progress.setProgress( i ); // update progress
            qApp->processEvents();     // as progress is a modal dialog, allow application to process events
            if ( progress.wasCancelled() ) {
                // user decided to stop
                f.close();
                for(j=0; j<nchan; j++) delete data[j];
                ndata = 0;
                return false;
            }
        }
    }

    updateData();

    return true;
}


void SpectrumExtractorWindow::updateData() {

    // create a new dataset to put the data into
    FSDataSet* ds = new FSDataSet(workingName, QPen(), 11, 1);
    ds->grow(ndata); // data copy will be quicker as no reallocation is necessary afterwards
    double caml  = display[0].caml;
    double delay = display[0].delay;
    double E0    = display[0].E0;

    long i;
    for(i=0; i<ndata; i++)  {
        double x = displayFunction(i, display[0].paramX);
        double y = displayFunction(i, display[0].paramY);
        ds->append(x,y);
    }

    if(caml!=0.) {
        // reshape time axis into mass
        for(i=0; i<ndata; i++)  {
            double invSpeed = (delay-ds->yp[i])/caml;
            ds->yp[i] = 2.*(ds->xp[i]+E0)*invSpeed*invSpeed;
        }
    }

    display[0].xmin = eventGraph->x->minC;
    display[0].xmax = eventGraph->x->maxC;
    display[0].ymin = eventGraph->y->minC;
    display[0].ymax = eventGraph->y->maxC;

    eventGraph->clearDataSets();
    eventGraph->appendDataSet(ds);

    // add polygons for selected elements
    int *selZ = periodicTable->multiSelected();
    if(selZ)
        for(i=1; i<=selZ[0]; i++)
            if(polygon[selZ[i]]) {
                FSDataSet *dsp = new FSDataSet(polygon[selZ[i]]);
                if(caml!=0.) {
                    // reshape polygon y coordinates
                    for(long j=0; j<dsp->getndata(); j++) {
                        double invSpeed = (delay-dsp->yp[j])/caml;
                        dsp->yp[j] = 2.*(dsp->xp[j]+E0)*invSpeed*invSpeed;
                    }
                }
                dsp->append(dsp->xp[0],dsp->yp[0]); // close the loop
                dsp->linePen = QPen(colorSeries(i),1); //selZ[i]),1);
                eventGraph->appendDataSet(dsp);
            }

    eventGraph->zoom(display[0].xmin, display[0].xmax, display[0].ymin, display[0].ymax);
}

void SpectrumExtractorWindow::selectChannel() {

    if(ndata==0) {
        QMessageBox::warning( this, "Spectrum Extractor", "Event file must be loaded to use this function.", QMessageBox::Ok, 0);
        return;
    }

    channelSelector * cs = new channelSelector(nchan, display[0], divide, specsize);
    if( cs->exec() ) {
        display[0] = cs->getDisplay();

        specsize = cs->spectrumSize->text().toLong();
        if(specsize<=0) specsize = 512;
        divide = cs->spectrumDivide->text().toLong();
        if(divide<=0) divide = 1;

        updateData();
    }
    delete cs;
}


void SpectrumExtractorWindow::Eprojection()
{
    projection(1);
}


// projecion of identified data
// type = 1: Energy projection using the E formula in the channel selector
// type = 2: Energy projection using time and reshape parameters to calculte the energy; y axis is assumed to be the mass
// type = 3: Progression: divide the event list in specsize intervals and sum the number of events for each identified atoms in each interval
void SpectrumExtractorWindow::projection(int type)
{
    // get selected atoms
    int *selZ = periodicTable->multiSelected();
    if(!selZ)
        return; // no element selected

    double caml  = display[0].caml;
    double delay = display[0].delay;
    double E0    = display[0].E0;

    // check parameters
    if(type<1 || type>3) {
        QMessageBox::warning( this, "Spectrum Extractor", "Unknown projection type", QMessageBox::Ok, 0);
        return;
    }
    if(type==2 && caml==0.) {
        QMessageBox::warning( this, "Spectrum Extractor", "Can't calculate energy from time without reshape", QMessageBox::Ok, 0);
        return;
    }

    // ask user if energyGraph should be cleared
    if(energyGraph->firstDataSet!=NULL &&
        QMessageBox::warning( this, "Spectrum Extractor",
                                    "Energy graph contains data.\n"
                                    "Should I Clear it?",
                                     QMessageBox::Yes | QMessageBox::Default,
                                     QMessageBox::No  | QMessageBox::Escape )==QMessageBox::Yes ) {
        energyGraph->clearDataSets();
    }

    long stepsize = ndata/specsize;  // bin size for progression (type 3)
    if(stepsize<1) stepsize = 1;

    for(long i=1; i<=selZ[0]; i++)
        // if a polygon is defined for the selected element ...
        if(polygon[selZ[i]]) {
            long j;
            Atom a = periodicTable->returnAtom(selZ[i]);

			if(type==2 && a.mass==0.) {
				QMessageBox::warning( this, "Spectrum Extractor", "Can't do time to energy conversion for dummy element " + a.name + ". Plotting time projection only.", QMessageBox::Ok, 0);
			}

            // create a histogram of length specsize, initializing bins to 0
            FSDataSet *dsp = new FSDataSet(polygon[selZ[i]]->legend,QPen(colorSeries(i),1)); //selZ[i]),1));
            for(j=0; j<specsize; j++)
                dsp->append(j,0);

            // create a copy of the current polygon...
            FSDataSet *polycopy = new FSDataSet(polygon[selZ[i]]);
            // ... and reshape its y coordinates if necessary
            if(caml!=0.) {
                for(j=0; j<polycopy->getndata(); j++) {
                    double invSpeed = (delay-polycopy->yp[j])/caml;
                    polycopy->yp[j] = 2.*(polycopy->xp[j]+E0)*invSpeed*invSpeed;
                }
            }

            // now the data...
            for(long k=0; k<ndata; k++)  {
                double x = displayFunction(k, display[0].paramX);
                double y = displayFunction(k, display[0].paramY);
                double deltat = delay-y; // in the case of type 2
                long E;

                // reshape data y coordinate
                if(caml!=0.) {
                    double invSpeed = (delay-y)/caml;
                    y = 2.*(x+E0)*invSpeed*invSpeed;
                }

                // is this point inside the polygon?
                if(pnpoly(polycopy->getndata(), polycopy->xp, polycopy->yp, x, y)) {
                    switch (type) {
                    case 1:
                        E = long(displayFunction(k, display[0].paramE)/divide); // energy for this event
                        if(E>0 && E<specsize) dsp->yp[E]++; // increment histogram
                        break;
                    case 2:
						E = long(deltat/divide);
                        if(E>0 && E<specsize) dsp->yp[E]++; // time histogram
                        break;
                    case 3:
                        dsp->yp[k/stepsize]++;  // sum the number of events in each interval
                        break;
                    default: // should not happen
                        break;
                    }
                }
            }

		    if(type==2 && a.mass!=0.) {

				// convert time axis into a non-linear E axis
				bool dataIgnored = false;
				for(j=0; j<specsize; j++) {
					double deltat = dsp->xp[j]*divide;  //multiply by "divide" so reshape formula applies
					dsp->xp[j] = deltat==0.?0.:(a.mass*caml*caml/deltat/deltat/2.-E0)/divide;

		// not used:// dN(E) = N(E)dE = N(E(t))*(dt/dE)*dE
					// where E = m*caml^2/(2(delay-t)^2)  and  dE/dt = (m*caml^2/2)/(E+E0)^(3/2)
					// dsp2->yp[j] *= (dsp2->xp[j]<=E0 ? 0. : pow(a.mass/2.,.5)*caml/pow(dsp2->xp[j]+E0,1.5));

					if( (dsp->xp[j]>specsize || dsp->xp[j]<0) && dsp->yp[j]!=0. ) dataIgnored=true;
				}

				if(dataIgnored)
					QMessageBox::warning( this, "Spectrum Extractor", "Time data correspond to energy outside specified energy spectrum. Data ignored.", QMessageBox::Ok, 0);

				dsp->sort();
				//dsp2->interpolate(specsize,dsp->xp,dsp->yp,0.);

				double * count = new double[specsize]; // number of particles on a linear E axis
				double S1x, S1y;
				double S2x, S2y;
				long is = 0;

				// find first channel with energy >= 0 in dsp
				for(j=0; j<specsize; j++)
					if(dsp->xp[j]>=0.) {
						S1x = dsp->xp[j];
						S1y = dsp->yp[j];
						S2x = dsp->xp[j+1];
						S2y = dsp->yp[j+1];
						is = j;
						break;
					} 

				long j1 = long(S1x);
				for(j=0; j<j1; j++) count[j]=0.;

				// redistribute the counts considering a linear E axis
				for(j=j1; j<specsize && is<specsize; j++) {

					// For the next calculation, it is assumed that the number of count S1y
				    // is the number of particules reaching the detector with S1x < E <S2x

				    // case where dE=(j+1)-j is so small that it is all contained in one interval S2x-S1x
					if(j+1<=S2x) {
						//double mm = (S2y-S1y)/(S2x-S1x);
						//double b = S1y-mm*S1x;
						//count[j] = mm*(j+.5)+b;  // linear interpolation between S1y and S2y
						count[j] = S1y/(S2x-S1x);
					}
						

					// case where dE=(j+1)-j covers 2 or more intervals
					else {
						count[j] = S1y*(S2x-j)/(S2x-S1x);
						while(S2x<j+1) {
							S1x = S2x;
							S1y = S2y;
							S2x = dsp->xp[is];
							S2y = dsp->yp[is];
							count[j] += S1y;
							is++;
						}
						count[j] -= S1y * (S2x-(j+1))/(S2x-S1x); // substract the exceeding count fraction
					}
				}				

				for(j=0; j<specsize; j++) {
					dsp->xp[j]=j;
					dsp->yp[j]=count[j];
				}

			}

	
			for(j=0; j<specsize; j++) {
                if(dsp->yp[j]>0) {
                    energyGraph->appendDataSet(dsp); // append the spectrum only if it contains events
                    break;
                }
            }
            if(j==specsize) delete dsp; // spectrum contains no events, so delete it

            qApp->processEvents(); // process system events, so energy spectra are displayed as they are appended
            //delete polycopy;  ** REMOVED BECAUSE IT GENERATES A DE-ALLOCATION PROBLEM (conflict between MSVC and Qt)
        }

    // if a formula is set, extract directly a spectrum from data[][]
    if(display[0].paramS[0]) {
        // create an histogram of length specsize, initializing bins to 0
        FSDataSet *dsSpec = new FSDataSet();
        long i;

        for(i=0; i<specsize; i++)
            dsSpec->append(i,0);

        // accumulate data in histogram
        for(long k=0; k<ndata; k++)  {
            long E = long(displayFunction(k, display[0].paramS)/divide); // energy for this event
            if(E>1 && E<specsize) dsSpec->yp[E]++; // increment histogram
        }

        // if the histogram contains events ...
        for(i=0; i<specsize; i++) {
            if(dsSpec->yp[i]>0) {
                // ...ask user for corresponding element name ...
                Atom atom = selectElement("Select element corresponding to spectrum channel");
                dsSpec->legend = atom.name;
                dsSpec->linePen = QPen(colorSeries(energyGraph->getndatasets()+1),1); //atom.Z),1);

                // ... and append spectrum to energy graph
                energyGraph->appendDataSet(dsSpec);
                break;
            }
        }
        if(i==specsize) delete dsSpec; // spectrum contains no events, so delete it
    }
}

// project identified data in energy spectra
void SpectrumExtractorWindow::Yprojection()
{

    if(ndata==0 || specsize<=0 || eventGraph->firstDataSet==NULL ) return;

    // make sure the selected dataset is the first (i.e. the event data)
    eventGraph->selectedDataSet = eventGraph->firstDataSet;

    // calculate the bin size so we end up with a dataset of size specsize
    double bin = (eventGraph->selectedDataSet->getyMax()-eventGraph->selectedDataSet->getyMin())/specsize;

    // projection (will contain only the data inside markers
    FSDataSet *dsp = eventGraph->yProjectionOfSelectedDataset(bin);
    if(dsp) {
        eventGraph->appendDataSet(dsp);
        dsp->swap();
    }

}

void SpectrumExtractorWindow::cancelReshape() {
    display[0].caml = 0.;
    eventGraph->unZoom();
    updateData();
}


Atom SpectrumExtractorWindow::selectElement(QString s)
{
    // Create a modal dialog window ...
    QDialog periodicTableDialog(this, "periodicTableDialog", true);
    periodicTableDialog.setFixedSize(QSize(500,255));
    periodicTableDialog.setCaption("[Qt] - " + s);

    // ...that will containt a periodic table widget to select a single element
    PeriodicTable periodicTableSelect(500,255, true, true, &periodicTableDialog);

    // when the widget tels that a element is selected, close the dialog
    connect(&periodicTableSelect, SIGNAL(elementSelected()), &periodicTableDialog, SLOT(accept()) );
    periodicTableDialog.exec();

    // return the selected element info
    return periodicTableSelect.selectedElement;

}

void SpectrumExtractorWindow::definePolygon() {

    long n = eventGraph->spots->getndata();
    if(n<3) {
        QMessageBox::warning( this, "Spectrum Extractor", "A polygon needs at least 3 marks to be defined.", QMessageBox::Ok, 0);
	return;
    }

    Atom atom = selectElement();
    int Z = atom.Z;
    double caml  = display[0].caml;
    double delay = display[0].delay;
    double E0    = display[0].E0;

    if(polygon[Z]) delete polygon[Z];
    polygon[Z] = new FSDataSet(atom.name,QPen(colorSeries(Z),1));
    //polygon[Z]->setsymbol(0);

    for(long i=0; i<n; i++) {
        double x = eventGraph->spots->xp[i];
        double y = eventGraph->spots->yp[i];
        if(caml!=0.)
            // graph is currently reshaped and we need de-shape time values (i.e. time and not mass)
            y = delay - pow(y*caml*caml/2/(x+E0),0.5);

        polygon[Z]->append(x,y);
    }

    eventGraph->spots->clear();

    updateData();

}


// This function returns true if datasets:
// - are present in the graph g
// - have all the same number of data
// - have all the same x coordinates
// It means that the data can be written in a multicolumn table (x, y1, y2, ...)
// and can be processed by the ERD program
bool SpectrumExtractorWindow::checkGraphData(FSGraph * g)
{
    // check if datasets present
    int ndataset = g->getndatasets();
    if(ndataset==0) {
        QMessageBox::warning( this, "Spectrum Extractor", "Graph contains no data.", QMessageBox::Ok, 0);
        return false;
    }

    // check if datasets have all the same set of x coordinates
    bool samex = true;
    double * x = g->firstDataSet->xp; // x coords of first datasets
    int ndata = g->firstDataSet->getndata();
    FSDataSet * ds = g->firstDataSet->nextDataSet;
    while(ds != NULL) { // check next dataset until there is no more
        if(ds->getndata()!=ndata) { // not the same number of data
            samex = false;
            break;
        }
        for(int i=0; i<ndata; i++) { // not the same x
            if(x[i]!=ds->xp[i]) {
                samex = false;
                break;
            }
        }
        if(!samex) break;
        ds = ds->nextDataSet;
    }
    if(!samex) {
        QMessageBox::warning( this, "Spectrum Extractor", "Data must have same x coordinates.", QMessageBox::Ok, 0);
        return false;
    }

    return true;

}


// write all spectra in eneryGraph into a multicolumn file (x, y1, y2, ...)
bool SpectrumExtractorWindow::writeMultiColumn()
{
    if(!checkGraphData(energyGraph))
        return false;


    // same x data, so lets ask for a file name and save
    QFileDialog fd;
    QString initFileName = workingDir + "/" + workingName + ".spec";
    QString fileName( fd.getSaveFileName(initFileName, "Allegria Spectrum Files (*.spec);;All Files (*.*)")  );
    if ( fileName.isNull() )
        return false;

    int slash = fileName.findRev(QRegExp("[\\/\\\\]"));
    if(slash!=-1) {
        workingDir = fileName.left( slash );
    }

    ofstream fo(fileName);
    FSDataSet * ds = energyGraph->firstDataSet;
    int n = ds->getndata();

    // write headers
    fo << "energy";
    FSDataSet * dt = ds;
    while(dt != NULL) {
        fo << " " << dt->getname();
        dt = dt->nextDataSet;
    }
    fo << "\n";

    // write the i_th spectrum value for each mass
    for(int i=0; i<n; i++) {
        fo << ds->xp[i];

        dt = ds;
        while(dt != NULL) {
            fo << " " << dt->yp[i];
            dt = dt->nextDataSet;
        }
        fo << "\n";
    }
    fo.close();


    return true;
}

// Write each spectrum in eneryGraph into separate files.
// File names have the form: workingName.legend.spc
// (legend must be 2 characters long: shorter length will be completed by "x",
//  longer length will be truncated)
// Ask user confirmation before overwriting a file
bool SpectrumExtractorWindow::writeMultiFile()
{
    if(energyGraph->firstDataSet==NULL) return false; // no datasets!

    // ask for a directory
    QString dirName( QFileDialog::getExistingDirectory ( workingDir, this, "savedir", "Select directory to save spectrum files") );
    if ( dirName.isNull() )
            return false;

    workingDir = dirName;

    FSDataSet * ds = energyGraph->firstDataSet;
    while(ds != NULL) {
        QString ext = ds->getname();
        switch (ext.length()) {
        case 0:
            ext = "xx";
            break;
        case 1:
            ext = ext+"x";
            break;
        case 2:
            // do nothing
            break;
        default: // all other lengths are truncated
            ext = ext.left(2);
        }

        QString fileName = workingDir + workingName.upper() + "." + ext.upper() + ".SPC";

        // check if we should (over)write this file
        bool doWrite = true;
        if(QFile::exists(fileName)) {
            switch (QMessageBox::warning( this, "Spectrum Extractor", "File " + workingName.upper() + "." + ext.upper() + ".SPC" + " already exist. Overwrite?",
                                                 "&Yes", "&No", "&Cancel", 0, 2 )  ) {
            case 0: //yes
                doWrite = true;
                break;
            case 1: // No
                doWrite = false;
                break;
            case 2: // cancel
            default:
                return false;
            }
        }

        if(doWrite) {
            ofstream fo(fileName);
            if(!fo) return false; // something went wrong

            long n = ds->getndata();
            for(long i=0; i<n; i++)
                fo << ds->xp[i] << " " << ds->yp[i] << "\n";

            fo.close();
            ds -> dataHaveChanged = false; // indicate to the dataset that it has been saved
        }
        ds = ds->nextDataSet;
    }

    return true;
}

// Write the event data corresponfing to each selected element in a separate file
// File names have the form: workingName.element.dat
// (legend must be 2 characters long: shorter length will be completed by "x",
//  longer length will be truncated)
// Ask user confirmation before overwriting a file
bool SpectrumExtractorWindow::writeElementEventFile()
{
    int *selZ = periodicTable->multiSelected();
    if(!selZ) return false; // no selected elements!

    // ask for a directory
    QString dirName( QFileDialog::getExistingDirectory ( workingDir, this, "savedir", "Select directory to save elements event files") );
    if ( dirName.isNull() )
        return false;

    workingDir = dirName;
    bool *isInside = new bool[ndata];
    double caml  = display[0].caml;
    double delay = display[0].delay;
    double E0    = display[0].E0;

    for(long i=1; i<=selZ[0]; i++) {
        QString ext = periodicTable->returnAtom(selZ[i]).name;

        switch (ext.length()) {
        case 0:
            ext = "xx";
            break;
        case 1:
            ext = ext+"x";
            break;
        case 2:
            // do nothing
            break;
        default: // all other lengths are truncated
            ext = ext.left(2);
        }

        QString fileName = workingDir + workingName + "." + ext + ".dat";

        // check if we sould (over)write a file
        bool doWrite = true;
        if(QFile::exists(fileName)) {
            switch (QMessageBox::warning( this, "Spectrum Extractor", "File " + workingName + "." + ext + ".dat" + " already exist. Overwrite?",
                                                 "&Ok", "&Cancel", 0, 2 )  ) {
            case 0: //yes
                doWrite = true;
                break;
            case 1: // No
                doWrite = false;
                break;
            case 2: // cancel
            default:
                return false;
            }
        }

        if(doWrite) {

            // create a copy of the current polygon...
            FSDataSet *polycopy = new FSDataSet(polygon[selZ[i]]);
            // ... and reshape its y coordinates if necessary
            if(caml!=0.) {
                for(long j=0; j<polycopy->getndata(); j++) {
                    double invSpeed = (delay-polycopy->yp[j])/caml;
                    polycopy->yp[j] = 2.*(polycopy->xp[j]+E0)*invSpeed*invSpeed;
                }
            }


            // for data that are inside the current polygon
            long k, nInside = 0;
            for(k=0; k<ndata; k++)  {
                double x = displayFunction(k, display[0].paramX);
                double y = displayFunction(k, display[0].paramY);

                // reshape data y coordinate
                if(caml!=0.) {
                    double invSpeed = (delay-y)/caml;
                    y = 2.*(x+E0)*invSpeed*invSpeed;
                }

                // is this point inside the polygon?
                if(pnpoly(polycopy->getndata(), polycopy->xp, polycopy->yp, x, y)) {
                    isInside[k] = true;
                    nInside++;
                } else {
                    isInside[k] = false;
                }
            }

            // now that we know the number of events inside the polygon, write them
            if(nInside>0) {
                ofstream fo(fileName);
                if(!fo) return false; // something went wrong

                fo << "Allegria event file" << "\n";
                fo << nInside << " " << nchan << "\n";
                for(k = 0; k<ndata; k++)
                    if(isInside[k]) {
                        for(long ch = 0; ch <nchan; ch++)
                            fo << data[ch][k] << " ";
                        fo << "\n";
                    }
                fo.close();
            }
        }
    }

    return true;

}



// copy display data from appropriate display[] element to current display (display[0])
void SpectrumExtractorWindow::recallDisplay(int n)
{
    if(n<NUMDISPLAY) {
        display[0] = display[n];

        // make sure recalled parameters do not point to channel numbers not available
        for(int i=1; i<NUMPARAM; i++) {
            if(display[0].paramX[i]>nchan) display[0].paramX[i] = 0;
            if(display[0].paramY[i]>nchan) display[0].paramY[i] = 0;
            if(display[0].paramE[i]>nchan) display[0].paramE[i] = 0;
            if(display[0].paramS[i]>nchan) display[0].paramS[i] = 0;
        }
        eventGraph->zoom(display[0].xmin, display[0].xmax, display[0].ymin, display[0].ymax);
        updateData();
    }

}

// store current display data (display[0]) into appropriate display[] element
void SpectrumExtractorWindow::storeDisplay(int n) {

    if(n>0 && n<NUMDISPLAY) {

        // first, update extrema values
        display[0].xmin = eventGraph->x->minC;
        display[0].xmax = eventGraph->x->maxC;
        display[0].ymin = eventGraph->y->minC;
        display[0].ymax = eventGraph->y->maxC;

        // ask for a name
        bool ok = FALSE;
        QString name = QInputDialog::getText( "Display name","Enter display name", QLineEdit::Normal ,display[n].name, &ok, this );
        if ( ok && !name.isEmpty() )
            display[0].name = name;
        else
            display[0].name = "display &"+QString::number(n);

        // then, store parameters
        display[n] = display[0];

        // update menu
	int i;
        recall->clear();
        store->clear();
        for(i=1; i<NUMDISPLAY; i++)
            recall->setItemParameter(recall->insertItem( display[i].name,  this, SLOT(recallDisplay(int))),i);

        for(i=1; i<NUMDISPLAY; i++)
            store->setItemParameter(store->insertItem( display[i].name,  this, SLOT(storeDisplay(int))),i);

    }
}


double SpectrumExtractorWindow::displayFunction(long i, long param[])
{

    // make sure we can access param[0...4]
    if(nparamerror && NUMPARAM<5) {
        QMessageBox::critical( this, "SpectrumExtractor", "NUMPARAM<5, Program will probably crash soon!", QMessageBox::Ok, 0);
        nparamerror = false;
    }

//  formula << "none" << "A+B+C+D" << "A+B-C-D" << "(A-B)/(C+D)" << "(A+B+C+D)/Ener";
    double div = 0.;

    switch(param[0]) {
    case 0:
        return 0.; // none
    case 1: // A+B+C+D
        return (param[1]>0?data[param[1]-1][i]:0.) + (param[2]>0?data[param[2]-1][i]:0.) + (param[3]>0?data[param[3]-1][i]:0.) + (param[4]>0?data[param[4]-1][i]:0.);
    case 2: // A+B-C-D
        return (param[1]>0?data[param[1]-1][i]:0.) + (param[2]>0?data[param[2]-1][i]:0.) - (param[3]>0?data[param[3]-1][i]:0.) - (param[4]>0?data[param[4]-1][i]:0.);
    case 3: // (A-B)/(C+D)
        div = (param[3]>0?data[param[3]-1][i]:0.) + (param[4]>0?data[param[4]-1][i]:0.);
        return div==0.?0.: ((param[1]>0?data[param[1]-1][i]:0.) - (param[2]>0?data[param[2]-1][i]:0.))/div;
    case 4: // (A+B+C+D)/Ener
        div = displayFunction(i,display[0].paramE);
        return div==0.?0.: ( (param[1]>0?data[param[1]-1][i]:0.) + (param[2]>0?data[param[2]-1][i]:0.) + (param[3]>0?data[param[3]-1][i]:0.) + (param[4]>0?data[param[4]-1][i]:0.) )/div;
    default:
        return 0.; // unknown formula
    }

}


void SpectrumExtractorWindow::about() {

QMessageBox::about ( this, "About SpectrumExtractor", 
"SpectrumExtractor (Allegria version 040513)\n (c) 2004 Franois Schiettekatte, Universit de Montral\n"
"Released under GNU General Public Licence (www.gnu.org). This program cannot be sold.\n\n"
"If you use SpectrumExtractor/Allegria for published work, please include the following citation:\n"
"F. Schiettekatte, M. Chicoine, S. Gujrathi, P. Wei, K. Oxorn, Allegria: a new interface to the ERD program, Nucl. Inst. Meth. B 219-220 (2004) 125");

}
