/* This file is part of the Allegria project
   Copyright (C) 2004 Francois Schiettekatte <francois.schiettekatte@umontreal.ca>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

// Dialog that displays a periodic table and that allows single or multiple element selection
// The table can be extended with dummy elements

#include <fstream.h>
#include <qmessagebox.h>
#include "PeriodicTable.h"

PeriodicTable::PeriodicTable(int w, int h, bool singleSelect, bool sshowDummy, QWidget *parent, QString name)
              : QWidget( parent, name )
{
	setCaption("Periodic Table");
	nAtoms = 0;

	showDummy = sshowDummy;

	loadAtoms();

    for(int i=1; i<nAtoms; i++) {
	    cb[i] = new QPushButton(Atoms[i].name,this );
		cb[i]->setToggleButton(true);

		if(singleSelect) 
			connect( cb[i], SIGNAL(clicked()), this, SLOT(singleElementSelect()));
		else
			connect( cb[i], SIGNAL(clicked()), this, SLOT(multiElementSelect()));
	}

    createTableOfElements(w,h);

	selectedElement = Atoms[0]; // set neutron as default
}

PeriodicTable::~PeriodicTable()
{
	for(int i=0; i<nAtoms; i++)
		if(Atoms[i].isotopes)
			delete Atoms[i].isotopes;

}

bool PeriodicTable::loadAtoms() {

	// load Atom info (name, most abundant mass, other isotope masses, etc.) 
	// from file alegria.ion
    char ch, name[100], abc[] = "abcdefghijklmnopqrstuvwxyz";
	nAtoms = 0;

    ifstream f("alegria.ion");
    if( !f ) {
		QMessageBox::warning( this, "AlegriaToF", "Error opening Atoms description file ALEGRIA.ION","OK");
        return false;
    }

    while(!f.eof())
    {
        f >> name >> Atoms[nAtoms].Z >> Atoms[nAtoms].mass >> Atoms[nAtoms].nIsotopes;
		Atoms[nAtoms].name = QString(name);
        Atoms[nAtoms].isotopes = new double[Atoms[nAtoms].nIsotopes];
        for(int i=0; i<Atoms[nAtoms].nIsotopes;i++)
            f >> Atoms[nAtoms].isotopes[i];

        f >> ch;
        if( ch!=';' ) {
			QMessageBox::warning( this, "AlegriaToF", "Error reading Atoms description file ALEGRIA.ION","OK");
            return false;
        }
        if(Atoms[nAtoms].name.length()>0)
            nAtoms++;
    }

    f.close();

	// add 16 dummy atoms
    for(int i=0; i<16; i++)
    {
        Atoms[nAtoms].Z=nAtoms;
		Atoms[nAtoms].mass = 0.;
		Atoms[nAtoms].nIsotopes = 0;
		Atoms[nAtoms].name = QString("Q") + QString(QChar(abc[i]));
        Atoms[nAtoms].isotopes = NULL;
		nAtoms++;
    }
	
	return true;


}

void PeriodicTable::createTableOfElements(int width, int height)
{

    int x, y;
	int bwidth = width/18;
	int bheight = int(height/(showDummy?11:9.2));

    // create standard periodic table of elements
    for(int i=1; i<nAtoms; i++) {
        int Z = Atoms[i].Z;

        // y coordinate (in button number) of element on table
        if(Z<3) y = 0;
        else if(Z<11) y = 1;
        else if(Z<19) y = 2;
        else if(Z<37) y = 3;
        else if(Z<55) y = 4;
        else if(Z>56 && Z<72) y = 7;
        else if(Z<87) y = 5;
        else if(Z>88 && Z<93) y = 8;
        else if(Z>=93) 
			y = 9;
        else y = 6;

        // x coordinate (in button number) of element on table
        switch (y) {
            case 0: x = (Z-1)*17; break;
            case 1: x = (Z- 3) + (Z>  4?10:0); break;
            case 2: x = (Z-11) + (Z> 12?10:0); break;
            case 3: x = (Z-19);                break;
            case 4: x = (Z-37);                break;
            case 5: x = (Z-55) - (Z> 71?14:0); break;
            case 6: x = (Z-87) - (Z>103?14:0); break;
            case 7: x = (Z-57)+2;              break;
            case 8: x = (Z-89)+2;              break;
            case 9: x = (Z-93);          break;
        }

        // position in pixel units
        x *= bwidth; //buttonSide;
        y *= bheight; //buttonSide; 

        // allow space between lanthanide/acinide series and rest of table
        if(y>6*bheight) y += int(bheight*.2);

        // allow space between lanthanide/acinide series and dummy elements
        if(y>9*bheight) y += int(bheight*.4);

        // create the button, a click event will call _buttonClick()
		cb[i]->move(QPoint(x,y));
		cb[i]->setFixedSize (bwidth,bheight);

    }
	lastClicked = 0;

    // resize window
	QRect geo = geometry();
	geo.setWidth(width);
	geo.setHeight(height);
	setGeometry(geo);

}

void PeriodicTable::singleElementSelect()
{
	// this fonction should be called only in case of singleSelect

	int i;

	if(lastClicked>0) cb[lastClicked]->setOn(false); // unclick previous selection
	lastClicked = 0;

    // check which element is selected: button is toggled on
	for (i=1; i<nAtoms; i++) {
		if(cb[i]->isOn())
				break;
	}

	if(i==nAtoms) // no button clicked!!
		selectedElement = Atoms[0]; // select neutron as default
	else {
		// set selected element to pushed button
		selectedElement = Atoms[i];
		emit elementSelected();
		lastClicked = i;
	}
}

void PeriodicTable::multiElementSelect()
{
		emit elementSelected();
}


Atom PeriodicTable::returnAtom(int atomNum) {
	if(atomNum>=1 || atomNum < nAtoms) 
		return Atoms[atomNum]; 

	return Atoms[0]; // return neutron if argument wrong
}


Atom PeriodicTable::returnAtom(QString atomName) {
	for (int i=1; i<nAtoms; i++) {
		if(atomName.upper() == Atoms[i].name.upper() || (Atoms[i].name.length()==1 && Atoms[i].name[0].upper()==atomName[0].upper()  ) )
			return Atoms[i];
	}

	return Atoms[0];  // return neutron if argument wrong
}

void PeriodicTable::setSelected(int Z, bool sel) {
	if(Z>=1 && Z < nAtoms)
		cb[Z]->setOn(sel); 

}


// returns a list of selected element in the periodic table.
// Index [0] contains the number of elements in the list
// If no element are selected, function returns NULL.
int *PeriodicTable::multiSelected()
{

	int i, n=0;

	// count the number of pressed buttons in element table
	for(i = 1; i< nAtoms; i++)
		if(cb[i]->isOn()) n++; 

	if(n==0) return NULL;

	// create a Z list of selected elements
	n = 1; // CAUTION: we start list at index 1
	for(i = 1; i< nAtoms; i++) 
		if(cb[i]->isOn()) ZSelected[n++]=i; 

	ZSelected[0] = n-1;

	return ZSelected;

}

void PeriodicTable::paintEvent( QPaintEvent * ) 
{
	QSize s = size();
	createTableOfElements(s.width(), s.height());

}

